package org.btcollider.cnc.keysrv.multistage;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;
import org.btcollider.cnc.dto.KeyRange;
import org.btcollider.cnc.keysrv.KeyServer;
import org.btcollider.cnc.keysrv.keytree.KeyTreeServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Not entirely random KeyServer that handles very large trees in multiple stages to reduce memory
 * and processing time consumption
 *
 * @author armin
 *
 */
public class MultiStagedServer implements KeyServer {
  final Logger log = LoggerFactory.getLogger(MultiStagedServer.class);
  private static final int PARTITION_DEPTH = 16;

  private KeyTreeServer primKTS;
  private KeyTreeServer secKTS;
  private KeyRange partition;
  private KeyRangeBuffer keyRangeBuffer;
  private int remainingDepth;
  private int subIndex;
  private long maxWorkSpan;

  public MultiStagedServer(int index, int depth, long maxWorkSpan, long pruneKey) {
    this.remainingDepth = depth - PARTITION_DEPTH;
    this.subIndex = index - PARTITION_DEPTH;
    this.maxWorkSpan = maxWorkSpan;

    // Build a reasonably sized KeyTreeServer with disabled KTConcierge service
    // This represents the first partitioning of the key space
    this.primKTS = new KeyTreeServer(index, PARTITION_DEPTH, 0, pruneKey);
    // Get a random starting range from the primary KTS, this range is still too huge
    this.partition = primKTS.getRange();
    // Start a secondary KTS to partition remaining key space in currentRange
    // This represents the actual work-horse KeyTreeServer for the current partition with
    // enabled KTConcierge service.
    // Needs subindex+1 because starting bit always set to 1
    long subPrune = ((long) 1 << subIndex + 1) | ((((long) 1 << subIndex + 1) - 1) & pruneKey);
    this.secKTS = new KeyTreeServer(subIndex + 1, remainingDepth, maxWorkSpan, subPrune);

    this.keyRangeBuffer = new KeyRangeBuffer(maxWorkSpan);
  }

  @Override
  public KeyRange getRange() {
    if (partition == null)
      return null; // We are finished searching the whole space

    KeyRange kr = null;

    if ((kr = keyRangeBuffer.getRange()) == null) { // Get retracted KR from buffer if one exists
      kr = secKTS.getRange(); // else: get a random range from the subspace
    }

    if (kr == null) { // sub-range finished
      log.info("Subrange {} finished. Opening new partition.", partition);
      primKTS.setSearched(partition);
      partition = primKTS.getRange();
      secKTS = new KeyTreeServer(subIndex + 1, remainingDepth, maxWorkSpan, 0);
      return getRange();
    }

    // strip out starting bit of subrange
    long krStart = kr.getStart() & ~(1 << subIndex + 1);
    // extract set tail bits
    long krEnd = kr.getStart() ^ kr.getEnd();

    // Combine partition with sub-range
    krStart = partition.getStart() | krStart;
    krEnd = krStart | krEnd;

    return new KeyRange(krStart, krEnd);
  }

  @Override
  public void setInWork(KeyRange keyRange) {
    keyRangeBuffer.setInWork(keyRange);

    // strip out partition bits -> get subspace KeyRange only
    long mask = ((long) 1 << subIndex + 1) - 1;
    // add leading bit
    long lb = ((long) 1 << subIndex + 1);

    KeyRange secKR =
        new KeyRange(((keyRange.getStart() & mask) | lb), ((keyRange.getEnd() & mask) | lb));

    secKTS.setInWork(secKR);
  }

  @Override
  public void setSearched(KeyRange keyRange) {

    keyRangeBuffer.setSearched(keyRange);
    // strip out partition bits -> get subspace KeyRange only
    long mask = ((long) 1 << subIndex + 1) - 1;
    // add leading bit
    long lb = ((long) 1 << subIndex + 1);

    KeyRange secKR =
        new KeyRange(((keyRange.getStart() & mask) | lb), ((keyRange.getEnd() & mask) | lb));

    secKTS.setSearched(secKR);
  }

  @Override
  public void dump(OutputStream out) throws IOException {
    synchronized (this) {
      ByteArrayOutputStream primDump = new ByteArrayOutputStream();
      primKTS.dump(primDump);
      primDump.flush();

      ByteArrayOutputStream secDump = new ByteArrayOutputStream();
      secKTS.dump(secDump);
      secDump.flush();

      out.write(primDump.toByteArray());
      out.write(0xFF);
      out.write(secDump.toByteArray());
      out.write(0xFF);
      out.flush();

      PrintWriter pw = new PrintWriter(new OutputStreamWriter(out, StandardCharsets.US_ASCII));
      pw.println();
      pw.println(partition.getStart() + "," + partition.getEnd());

      keyRangeBuffer.getAllRanges().forEach(kr -> pw.println(kr.getStart() + "," + kr.getEnd()));
      pw.flush();

      out.flush();
    }
  }

  @Override
  public void undump(InputStream in) throws IOException {
    synchronized (this) {
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      for (int b = in.read(); b != 0xFF; b = in.read()) {
        bos.write(b);
      }
      ByteArrayInputStream bis = new ByteArrayInputStream(bos.toByteArray());
      primKTS.undump(bis);

      bos = new ByteArrayOutputStream();
      for (int b = in.read(); b != 0xFF; b = in.read()) {
        bos.write(b);
      }
      bis = new ByteArrayInputStream(bos.toByteArray());
      secKTS.undump(bis);

      BufferedReader br = new BufferedReader(new InputStreamReader(in, StandardCharsets.US_ASCII));
      br.readLine();
      String[] partitionDump = br.readLine().split(",");
      this.partition =
          new KeyRange(Long.parseLong(partitionDump[0]), Long.parseLong(partitionDump[1]));

      String krbuf;
      while ((krbuf = br.readLine()) != null) {
        String[] krbufsplit = krbuf.split(",");
        keyRangeBuffer
            .setInWork(new KeyRange(Long.parseLong(krbufsplit[0]), Long.parseLong(krbufsplit[1])));
      }
    }
  }

}
